@@ -31,7 +31,7 @@ extern "C" {
 #endif
 #endif
 
-#define GSASL_VERSION "0.0.0"
+#define GSASL_VERSION "0.0.1"
 
 /* Quoting RFC 2222:
    SASL mechanisms are named by strings, from 1 to 20 characters in
@@ -86,6 +86,7 @@ enum {
   GSASL_MECHANISM_PARSE_ERROR,
   GSASL_AUTHENTICATION_ERROR,
   GSASL_CANNOT_GET_CTX,
+  GSASL_INTEGRITY_ERROR,
   GSASL_NO_MORE_REALMS
 };
 
@@ -155,6 +156,14 @@ extern int gsasl_server_step_base64 (Gsasl_session_ctx *xctx,
 				     size_t b64output_len);
 extern void gsasl_server_finish (Gsasl_session_ctx *xctx);
 
+/* Session functions: xcode.c */
+extern int gsasl_encode (Gsasl_session_ctx *xctx,
+			 const char *input, size_t input_len,
+			 char *output, size_t *output_len);
+extern int gsasl_decode (Gsasl_session_ctx *xctx, 
+			 const char *input, size_t input_len,
+			 char *output, size_t *output_len);
+
 /* Error handling: error.c */
 extern const char *gsasl_strerror (int err);
 
