@@ -40,6 +40,13 @@ extern char *_gsasl_gettext (const char *str);
 #define HEXCHAR(c) ((c & 0x0F) > 9 ? 'a' + (c & 0x0F) - 10 : '0' + (c & 0x0F))
 #define MIN(a,b) ((a) < (b) ? (a) : (b))
 
+typedef int (*_Gsasl_code_function) (Gsasl_session_ctx *cctx,
+				     void *mech_data,
+				     const char *input,
+				     size_t input_len,
+				     char *output,
+				     size_t *output_len);
+
 struct _Gsasl_mechanism_functions {
   int	(*init)   (Gsasl_ctx *ctx);
   void	(*done)   (Gsasl_ctx *ctx);
@@ -53,6 +60,8 @@ struct _Gsasl_mechanism_functions {
 		   size_t *output_len);
   int	(*finish) (Gsasl_session_ctx *cctx, 
 		   void *mech_data);
+  _Gsasl_code_function encode;
+  _Gsasl_code_function decode;
 };
 
 struct _Gsasl_mechanism {
@@ -97,6 +106,7 @@ struct Gsasl_ctx {
 
 struct Gsasl_session_ctx {
   Gsasl_ctx *ctx;
+  int clientp;
   _Gsasl_mechanism *mech;
   void *application_data;
   void *mech_data;
